# -*- coding: utf-8 -*-

import sqlite3
from datetime import datetime

#@@FIXME: Use app.DATABASE_PATH
DATABASE_PATH = './data/bikini.db'

def connect(database=DATABASE_PATH):
    conn = sqlite3.connect(database, detect_types=sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES)        
    conn.row_factory = sqlite3.Row    
    # "Foreign key constraints are disabled by default (...), so must be   
    #   enabled separately for each database connection separately"
    conn.execute('PRAGMA foreign_keys = ON')
    return conn

def add_page(conn, name, title, content_html, content_hash):
    conn.execute('''
        INSERT INTO pages(name, title, content_html, content_hash) 
        VALUES (?, ?, ?, ?)''', 
        (name, title, content_html, content_hash))            

# def remove_page(conn, name):
#     pass

def update_page(conn, name, title, content_html, content_hash, is_deleted=False):    
    conn.execute('''
        UPDATE pages SET title = ?, content_html = ?, content_hash = ?, is_deleted = ?
        WHERE name = ?''', 
        (title, content_html, content_hash, is_deleted, name))                 


def add_revision(conn, name, revision_id, content, author, host_name, message=''):    
    conn.execute('''
        INSERT INTO revisions(revision_id, content, author, host_name, message, page_name) VALUES 
        (?, ?, ?, ?, ?, ?)
        ''', (revision_id, content, author, host_name, message, name))        

def archive_revision(conn, name, revision_id):    
    conn.execute('''
        UPDATE revisions SET archived_on = CURRENT_TIMESTAMP 
        WHERE page_name = ? and revision_id = ?''', 
        (name, revision_id))       

def purge_revisions(conn, days_limit):    
    #@@TODO: if days_limit ...    
    query = '''DELETE FROM revisions  
        WHERE archived_on NOT NULL and archived_on < datetime('now', '-%d days')''' % days_limit        
    c = conn.cursor()    
    c.execute(query)       
    
    return c.rowcount

def purge_deleted_pages(conn, days_limit):    
    query = '''DELETE FROM pages  
        WHERE is_deleted = 1 and archived_on < datetime('now', '-%d days')''' % days_limit        
    c = conn.cursor()    
    c.execute(query)       
    
    return c.rowcount
    
def get_page(conn, name):    
    cursor = conn.execute('''
        SELECT pages.*, 
        revisions.revision_id, 
        revisions.content,                 
        revisions.created_on AS "last_updated_on [timestamp]", 
        revisions.author AS "last_author" FROM pages 
        INNER JOIN revisions ON name = ? AND revisions.page_name = pages.name 
        ORDER BY revisions.created_on DESC LIMIT 1''', (name, ))    
    return cursor.fetchone()
    
def get_page_revision(conn, name, revision_id):
    cursor = conn.execute('''
        SELECT * FROM pages INNER JOIN revisions 
        ON name = ? AND revisions.page_name = pages.name AND revision_id = ?''', (name, revision_id))    
    return cursor.fetchone()

# def rollback_revision(conn, name, revision_id):
#     raise NotImplementedError


def update_links(conn, name, links):
    
    with conn:    
        conn.execute("DELETE FROM links WHERE page_name = ?", (name, ))
        conn.executemany('''INSERT OR REPLACE INTO links (page_name, linked_page_name) VALUES (?, ?)''', 
            [(name, link) for link in links])                 


def get_inbound_links(conn, name):
    cursor = conn.execute('''
        SELECT page_name FROM links WHERE linked_page_name = ?''', (name, ))    
    return [row['page_name'] for row in cursor]

def get_outbound_links(conn, name):
    cursor = conn.execute('''
        SELECT DISTINCT linked_page_name FROM links WHERE page_name LIKE ?''', (name, ))    
    return [row['linked_page_name'] for row in cursor]
    
def get_all_names(conn):
    cursor = conn.execute("SELECT name FROM pages ORDER BY name")        
    return [row['name'] for row in cursor]

def get_recent_changes(conn, days_limit=0):
    if days_limit:
        query = '''SELECT * FROM pages INNER JOIN revisions 
        ON revisions.page_name = pages.name 
        WHERE created_on > datetime('now', '-%d days') ORDER BY created_on DESC''' % days_limit    
    else:
        query = '''SELECT * FROM pages INNER JOIN revisions 
        ON revisions.page_name = pages.name 
        ORDER BY created_on DESC'''
        
    cursor = conn.execute(query)            
    return cursor

def get_page_revisions(conn, name, days_limit=0):
    if days_limit:
        query = '''SELECT * FROM pages INNER JOIN revisions 
        ON name = ? AND revisions.page_name = pages.name 
        WHERE created_on > datetime('now', '-%d days') ORDER BY created_on DESC''' % days_limit
    else:
        query = '''SELECT * FROM pages INNER JOIN revisions 
        ON name = ? AND revisions.page_name = pages.name ORDER BY created_on DESC'''

    cursor = conn.execute(query, (name, ))    
    return cursor

def setup_db(conn):
    '''
    Setup database structure.
    '''
    with conn:
        conn.execute('''
            CREATE TABLE IF NOT EXISTS pages (
                name TEXT UNIQUE,             
                title TEXT,
                content_html TEXT,
                content_hash TEXT,
                is_deleted INTEGER DEFAULT 0
            )'''
        )            
        conn.execute('''
            CREATE TABLE IF NOT EXISTS revisions (
                revision_id INTEGER DEFAULT 1, 
                content TEXT, 
                created_on timestamp DEFAULT CURRENT_TIMESTAMP, 
                archived_on timestamp,
                author, 
                host_name,
                message TEXT,
                page_name TEXT,            
                FOREIGN KEY(page_name) REFERENCES pages(name) ON DELETE CASCADE
            )'''
        )
        conn.execute("CREATE INDEX IF NOT EXISTS idx_revision_id ON revisions (revision_id)")        
        conn.execute('''
            CREATE TABLE IF NOT EXISTS links (
                page_name TEXT,
                linked_page_name TEXT,
                FOREIGN KEY(page_name) REFERENCES pages(name) ON DELETE CASCADE
            )'''
        )
        conn.execute("CREATE UNIQUE INDEX IF NOT EXISTS idx_page_name_linked_page_name ON links (page_name, linked_page_name)")
        

# ----------------------------------------------
# Tests
# ----------------------------------------------

import random
from datetime import datetime

def test_fill_db(conn):
    
    content_lorem = "Sit Pharetra Porta Dolor Bibendum. Justo Venenatis Sem Purus Elit. Justo Aenean Ultricies Magna".split()    
    name_ipsum = "FrontPage Risus Sem Porta Magna Adipiscing Vehicula Mollis Bibendum Ligula Euismod".split()
    now = datetime.now()
    
    for name in name_ipsum:        
        random.shuffle(content_lorem)
        with conn:
            add_page(conn, name, name, ' '.join(content_lorem), 'somehash')
            add_revision(conn, name, 1, ' '.join(content_lorem),  'AnonymousCoward',  '127.0.0.1')

def run_tests():
    conn = connect(database=':memory:')

    setup_db(conn)
    test_fill_db(conn)
    row = get_page(conn, 'Adipiscing')
    assert row['name'] == 'Adipiscing'
    assert row['title'] == 'Adipiscing'
    assert row['last_author'] == 'AnonymousCoward'
    assert type(row['last_updated_on']) is datetime
    
    row = get_page(conn, 'xxx')
    assert row == None

    with conn:
        update_page(conn, 'Adipiscing', 'Long title for Adipiscing', 'New HTML', 'otherhash')
        add_revision(conn, 'Adipiscing', 2, 'New Content', 'YetOtherAnonymous', 'hostname.address.net')
        update_links(conn, 'Adipiscing', ['Foo', 'Foo', 'Bar'])

    row = get_page_revision(conn, 'Adipiscing', 2)
    assert row and row['revision_id'] == 2
    assert 'Adipiscing' in get_inbound_links(conn, 'Foo')        
    assert 'Foo' in get_outbound_links(conn, 'Adipiscing')

    with conn:
        archive_revision(conn, 'Adipiscing', 2) 
        add_revision(conn, 'Adipiscing', 3, 'Other Content', 'AnonymousCoward', 'example.com')
    row = get_page_revision(conn, 'Adipiscing', 2)
    assert row and row['archived_on'] is not None


    conn.close()
    
    
if __name__ == '__main__':
    run_tests()